-- Copyright Jeff Arneson 2012
-- Last Modified: 16 January 2012
-- View readme for everything else.

STRICT = true
DEBUG = true

require 'zoetrope'
require 'TEsound'

Unit = Animation:extend
{
	canJump = false,
	hasFocus = false,
	lastElapsed = 0,
	frozeness = 0,
	
	onCollide = function (self, other, h, v)
		if other:instanceOf(RedPortal) or other:instanceOf(Switch) or other:instanceOf(Vault) or other:instanceOf(Sign)then
			--nothing
		elseif other:instanceOf(FreezeRayBlast) then
			self.frozeness = math.min(self.frozeness + 5, 100)
			TEsound.play('wav/ice-beam_hit.wav')
		elseif other:instanceOf(FloorButton) then
			if (v >= 26) and self.velocity.y > 0 then
				self.velocity.y = 0
				self.acceleration.y = 0
				if (self.y < other.y) then
					self.y = math.floor(self.y) - (v - 26)
				end
				self.canJump = true
			end
		elseif other:instanceOf(Box) then
			if (v > h) then
				if (self.x < other.x) then
					self.x = self.x - h
				else
					self.x = self.x + h
				end
		
				if self.additionalHorizontalCollisions and v > 4 then
					self:additionalHorizontalCollisions(other, h, v)
				end
			elseif self.y < other.y then
				self.velocity.y = 0
				self.acceleration.y = 0
				self.y = self.y - v
				self.canJump = true
			end
		else
			local error = math.abs(math.abs(self.velocity.x * self.lastElapsed) - h - 1)
			if h < v or ( error < 0.05 ) then
				self.velocity.x = 0
				if (self.x < other.x) then
					self.x = self.x - h
				else
					self.x = self.x + h
				end
				
		
				if self.additionalHorizontalCollisions and v > 4 then
					self:additionalHorizontalCollisions(other, h, v)
				end
			else
				self.velocity.y = 0
				self.acceleration.y = 0
				if (self.y < other.y) then
					self.y = math.floor(self.y) - v + 1
					self.canJump = true
				else
					self.y = math.ceil(self.y) + v 
				end
			end
		end
	end
}

Player = Unit:extend
{
	width = 32,
	height = 64,
	image = 'img/Doc.png',
	normalAcceleration = { x = 0, y = 500},
	lastElapsed = 0,
	facing = 0,
	sequences = 
	{
		idle = { frames = {1, 2}, fps = 3},
		right = { frames = {11, 12, 13, 14, 15, 11, 16, 17, 18, 19}, fps = 14},
		left = { frames = {21, 22, 23, 24, 25, 21, 26, 27, 28, 29}, fps = 14}
	},
	
	runCollisions = function (self)
		the.view.mapView.background:subcollide(self)
		the.view.boxes:collide(self)
		the.view.doors:collide(self)
		the.view.buttons:collide(self)
	end,
	
	onUpdate = function (self, elapsed)
		self.lastElapsed = elapsed
		if the.keys:pressed('a') and self.hasFocus then
			self.velocity.x = -150
			self:play('left')
			self.facing = -1
		elseif the.keys:pressed('d') and self.hasFocus then
			self.velocity.x = 150
			self:play('right')
			self.facing = 1
		else
			self.velocity.x = 0
			self:play('idle')
			self.facing = 0
		end
		if the.keys:pressed('w') and self.hasFocus and self.canJump then
			self.y = self.y - 1
			self.velocity.y = -200
			TEsound.play('wav/doc_jump.wav')
		end
		
		if the.keys:justPressed(' ') and self.hasFocus then
			the.view.switches:collide(self)
			the.view.vaults:collide(self)
			the.view.signs:collide(self)
		end
		
		self.canJump = false
		self:runCollisions()
		self.acceleration.x = self.normalAcceleration.x
		self.acceleration.y = self.normalAcceleration.y
	end,
}

Cat = Unit:extend
{
	width = 32,
	height = 32,
	image = 'img/cat.png',
	sequences = 
	{
		idle = { frames = {1, 2, 3, 2}, fps = 6},
		right = { frames = {11, 12, 13, 14, 15}, fps = 12},
		left =  { frames = {21, 22, 23, 24, 25}, fps = 12}
	},
	normalAcceleration = { x = 0, y = 500},
	lastElapsed = 0,
	facing = 0,
	
	runCollisions = function (self)
		the.view.mapView.background:subcollide(self)
		the.view.boxes:collide(self)
		the.view.doors:collide(self)
		the.view.buttons:collide(self)
	end,
	
	onUpdate = function (self, elapsed)
		self.lastElapsed = elapsed
		if the.keys:pressed('a') and self.hasFocus then
			self.velocity.x = -250
			self:play('left')
			self.facing = -1
		elseif the.keys:pressed('d') and self.hasFocus then
			self.velocity.x = 250
			self:play('right')
			self.facing = 1
		else
			self.velocity.x = 0
			self:play('idle')
			self.facing = 0
		end
		if the.keys:pressed('w') and self.hasFocus and self.canJump then
			self.y = self.y - 1
			self.velocity.y = -420
			TEsound.play('wav/cat_jump.wav')
		end
		
		if the.keys:justPressed(' ') and self.hasFocus then
			the.view.switches:collide(self)
		end
		
		self.canJump = false
		self:runCollisions()
		self.acceleration.x = self.normalAcceleration.x
		self.acceleration.y = self.normalAcceleration.y
	end,
}

Person = Unit:extend
{
	width = 32,
	height = 64,
	image = 'img/people.png',
	normalAcceleration = { x = 0, y = 500},
	lastElapsed = 0,
	facing = 0,
	directionChangeEvery = 5.0,
	lastDirectionChange = 5.0,
	
	
	onNew = function (self)
		local d = math.random(0, 3)
		self.sequences = 
		{
			idle = { frames = {1 + 2*d, 2 + 2*d}, fps = 3},
			right = { frames = {20*d + 11, 20*d + 12, 20*d + 13, 20*d + 14, 20*d + 15, 20*d + 11, 20*d + 16, 20*d + 17, 20*d + 18, 20*d + 19}, fps = 14},
			left = { frames = {20*d + 21, 20*d + 22, 20*d + 23, 20*d + 24, 20*d + 25, 20*d + 21, 20*d + 26, 20*d + 27, 20*d + 28, 20*d + 29}, fps = 14}
		}
	end,
	
	runCollisions = function (self)
		the.view.mapView.background:subcollide(self)
		the.view.boxes:collide(self)
		the.view.doors:collide(self)
		the.view.buttons:collide(self)
	end,
	
	onUpdate = function (self, elapsed)
		self.lastElapsed = elapsed
		if self.frozeness == 0 then
			self.lastDirectionChange = self.lastDirectionChange + elapsed
			
			
			if self.lastDirectionChange >= self.directionChangeEvery then
				self.facing = math.random(-1, 1)
				self.lastDirectionChange = 0
				if self.facing == 0 then
					self.lastDirectionChange = self.directionChangeEvery - 1
				end
			end
				if self.facing == 1 then
					self.velocity.x = 150
					self:play('right')
				elseif self.facing == -1 then
					self.velocity.x = -150
					self:play('left')
				else
					self.velocity.x = 0
					self:play('idle')
				end
		else
			self.velocity.x = 0
			self:freeze()
		end
		
		self.canJump = false
		self:runCollisions()
		self.acceleration.x = self.normalAcceleration.x
		self.acceleration.y = self.normalAcceleration.y
		
		self.frozeness = math.max(self.frozeness - elapsed * 8, 0)
		
		self.tint[1] = math.min(math.max(1 - (self.frozeness / 100),0), 1)
		self.tint[2] = math.min(math.max(1 - (self.frozeness / 100),0), 1)
		self.tint[3] = math.min(math.max(1,0), 1)
	end,
	
	additionalHorizontalCollisions = function (self, other,h, v)
		self.lastDirectionChange = self.directionChangeEvery
	end
}

PersonSpawner = Fill:extend
{
	width = 32,
	height = 32,
	fill = {255, 255, 255, 100},
	border = {255, 255, 255, 255},
	
	onNew = function (self)
		self.MyPerson = Person:new{x = self.x, y = self.y}
		the.view.people:add(self.MyPerson)
	end
}

Box = Tile:extend
{
	image = 'img/tileset.png',
	imageOffset = {x = 0, y = 32},
	normalAcceleration = { x = 0, y = 100},
	lastElapsed = 0,
	collisions = 0,
	horizontalCollision = false,
	drag = {x = 200, y = 0},
	direction = 0,
	inHands = false,
	
	
	runCollisions = function(self)
		the.view.mapView.background:subcollide(self)
		the.view.doors:collide(self)
		the.view.boxes:collide(self)
	end,
	
	onUpdate = function(self, elapsed)
		--self.velocity.x = self.velocity.x * .98
		self.direction = 0
		--self.tint = {1, 1, 1}
		self.lastElapsed = elapsed
		self.horizontalCollision = false
		--self.tint = {1, 1, 1}
		self:runCollisions()
		self.acceleration.x = self.normalAcceleration.x
		self.acceleration.y = self.normalAcceleration.y
	end,
	
	onEndFrame = function(self)
		--self.velocity.x = 0
	end,
	
	onCollide = function(self, other, h, v)
		--self.tint = {0, 1, 0}
		if other:instanceOf(Unit) then
		elseif other:instanceOf(FloorButton) then
			if (v >= 26) then
				self.velocity.y = 0
				self.acceleration.y = 0
				if (self.y < other.y) then
					self.y = math.ceil(self.y) - (v - 26)
				end
			end
		elseif other:instanceOf(Box) then
			if not other.inHands then 
				if (v > h) then
					if (self.x < other.x) then
						self.x = self.x - h
					else
						self.x = self.x + h
					end
					self.horizontalCollision = true
				elseif self.y < other.y then
					self.velocity.y = 0
					self.acceleration.y = 0
					self.y = math.ceil(self.y) - v
				end
			end
		else
			if (v > h) then
				if (self.x < other.x) then
					self.x = self.x - h
				else
					self.x = self.x + h
				end
				self.horizontalCollision = true
			else
				self.velocity.y = 0
				self.acceleration.y = 0
				if (self.y < other.y) then
					self.y = math.ceil(self.y) - v
				else
					self.y = math.ceil(self.y) + v
				end
			end
		end
	end
}

FloorButton = Animation:extend
{
	image = 'img/button.png',
	sequences = 
	{
		up = { frames = {1}, fps = 1 },
		down = { frames = {2}, fps = 1 }
	},
	width = 32,
	height = 32,
	pushed = false,
	downSoundPlayed = false,
	upSoundPlayed = true,
	
	onUpdate = function (self, elapsed)
		if self.pushed then
			if not self.downSoundPlayed then
				self.downSoundPlayed = true
				if self:distanceTo(the.player) <= 550 then
					TEsound.play('wav/button_click.wav')
				end
			end
			self:play('down')
			self.upSoundPlayed = false
		else
			if not self.upSoundPlayed then
				self.upSoundPlayed = true
				if self:distanceTo(the.player) <= 550 then
					TEsound.play('wav/button_click.wav')
				end
			end
			self:play('up')
			self.downSoundPlayed = false
		end
	end,
	
	onEndFrame = function (self)
		self.pushed = false
	end,
	
	onCollide = function (self, other, h, v)
		if other:instanceOf(Unit) or other:instanceOf(Box) then
			if v >= 24 then
				self.pushed = true
			end
		end
	end
}

Switch = Tile:extend
{
	image = 'img/tileset.png',
	imageOffset = { x = 32, y = 32},
	width = 32,
	height = 32,
	solid = true,
	flipped = false,
	
	onCollide = function (self, other, h, v)
		self.flipX = not self.flipX
		self.flipped = not self.flipped		
	end
}

Door = Animation:extend
{
	width = 32,
	height = 32,
	solid = true,
	opened = false,
	openedSoundPlayed = false,
	closedSoundPlayed = true,
	
	open = function (self)
		if not self.openedSoundPlayed then
			self.openedSoundPlayed = true
			if self:distanceTo(the.player) <= 550 then
				TEsound.play('wav/door_toggle.wav', 'doorsw', 0.2)
			end
		end
		self.opened = true
		self.solid = false
		self:play('open')
		self.closedSoundPlayed = false
	end,
	
	close = function (self)
		if not self.closedSoundPlayed then
			self.closedSoundPlayed = true
			if self:distanceTo(the.player) <= 550 then
				TEsound.play('wav/door_toggle.wav', 'doorsw', 0.2)
			end
		end
		self.opened = false
		self.solid = true
		self:play('closed')
		self.openedSoundPlayed = false
	end,
	
	onUpdate = function (self, elapsed)
		local shouldBeOpen = true
		local switches = false
		
		if not self.switchGroups then
			switches = true
		end
		local x = 0
		if self.group then
			for _, spr in pairs(the.view.buttons.sprites) do
				if self.group == spr.group then
					if not spr.pushed then
						shouldBeOpen = false
					end
					x = x + 1
				end
			end
			for _, spr in pairs(the.view.switches.sprites) do
				if self.switchGroups then
					local groups = {}
					for g in string.gmatch(self.switchGroups, '%w+') do
						if g == spr.group then
							switches = xor(switches, spr.flipped)
						end
					end
				elseif self.group == spr.group then
					shouldBeOpen = shouldBeOpen and spr.flipped
				end
			end
		end
		
		if shouldBeOpen and switches then
			self:open()
		else
			self:close()
		end
	end
}

DoorTop = Door:extend
{
	image = 'img/tileset.png',
	sequences = 
	{
		open = { frames = {22}, fps = 1},
		closed = { frames = {21}, fps = 1}
	},
}

DoorBottom = Door:extend
{
	image = 'img/tileset.png',
	sequences = 
	{
		open = { frames = {32}, fps = 1},
		closed = { frames = {31}, fps = 1}
	},
}

DoorLeft = Door:extend
{
	image = 'img/tileset.png',
	sequences = 
	{
		open = { frames = {33}, fps = 1},
		closed = { frames = {23}, fps = 1}
	},
}

DoorRight = Door:extend
{
	image = 'img/tileset.png',
	sequences = 
	{
		open = { frames = {34}, fps = 1},
		closed = { frames = {24}, fps = 1}
	},
}

xor = function(a, b)
	x = ((a and not b) or (b and not a))
	
	return x
end

Vault = Animation:extend
{
	image = 'img/bigvault.png',
	sequences = 
	{
		open = { frames = {2}, fps = 1},
		closed = { frames = {1}, fps = 1},
	},
	height = 96,
	width = 64,
	Opened = false,
	openSoundPlayed = false,
	
	onUpdate = function(self, elapsed)
		if self.Opened then
			if not self.openSoundPlayed then
				TEsound.play('wav/vault_open.wav')
				self.openSoundPlayed = true
			end
			self:play('open')
		else
			self:play('closed')
		end
	end,
	
	onCollide = function(self, other, h, v)
		self.Opened = true
	end
}

SignView = Subview:extend
{
	onNew = function(self)
		self.background = Fill:new
		{
			x = 64,
			y = 64,
			width = 800 - 128,
			height = 300 - 64,
			fill = {0, 0, 0, 150}
		}
		
		self.text = Text:new
		{
			x = 66,
			y = 66,
			width = 800 - 128 - 2,
			height = 300 - 64 - 10,
			text = '...',
			wordWrap = true,
			font = 14
		}
		
		self.exitText = Text:new
		{
			x = 66,
			y = 300 - 12,
			width = 800,
			height = 10,
			text = 'Esc to Close',
			font = 10
		}
		
		self:add(self.background)
		self:add(self.exitText)
		self:add(self.text)		
	end,
	
	onUpdate = function(self, elapsed)
		if the.keys:justPressed('escape') then
			self:deactivate()
		end
	end
}

Sign = Tile:extend
{
	image = 'img/tileset.png',
	imageOffset = { x = 64, y = 32},
	height = 32,
	width = 32,
	
	onNew = function (self)
		self.Subview = SignView:new{}
		if self.signText then
			self.Subview.text.text = self.signText
		end
	end,
	
	onCollide = function (self, other, h, v)
		self.Subview:activate()
	end
}

DocGadget = Class:extend
{
	Icon = '',
	GadgetName = '',
	
	LeftMouseClick = function(self)
		if self.onLeftMouseClick then
			local xpos = the.mouse.x - the.view.translate.x
			local ypos = the.mouse.y - the.view.translate.y
			self:onLeftMouseClick(xpos, ypos)
		end
	end,
	
	LeftMouseUp = function(self)
		if self.onLeftMouseUp then
			local xpos = the.mouse.x - the.view.translate.x
			local ypos = the.mouse.y - the.view.translate.y
			self:onLeftMouseUp(xpos, ypos)
		end
	end,
	
	RightMouseClick = function(self)
		if self.onRightMouseClick then
			local xpos = the.mouse.x - the.view.translate.x
			local ypos = the.mouse.y - the.view.translate.y
			self:onRightMouseClick(xpos, ypos)
		end
	end,
	
	RightMouseUp = function(self)
		if self.onRightMouseUp then
			local xpos = the.mouse.x - the.view.translate.x
			local ypos = the.mouse.y - the.view.translate.y
			self:onRightMouseUp(xpos, ypos)
		end
	end,
	
	FrameListener = function(self, elapsed)
		if self.onUpdate then
			local xpos = the.mouse.x - the.view.translate.x
			local ypos = the.mouse.y - the.view.translate.y
			self:onUpdate(xpos, ypos, elapsed)
		end
	end
}

RobotArms = DocGadget:extend
{
	Icon = 'img/robot-arms.png',
	GadgetName = 'Robot Arms',
	
	onLeftMouseClick = function(self, xpos, ypos)
		if self.SelectedBox then
			if the.player.facing == 0 then
				self.SelectedBox.y = the.player.y + the.player.height / 2 - 1
			end
			self.SelectedBox.solid = true
			if not the.view.mapView.background:subcollide(self.SelectedBox) and not the.view.boxes:collide(self.SelectedBox) then
				self.SelectedBox.inHands = false
				self.SelectedBox.alpha = 1
				self.SelectedBox = nil
				TEsound.play('wav/box_grab.wav', '', 0.6)
			else
				self.SelectedBox.solid = false
			end
		else
		-- find a box under the mouse
			for _, spr in pairs(the.view.boxes.sprites) do
				if xpos >= spr.x and xpos <= spr.x + spr.width and ypos >= spr.y and ypos <= spr.y + spr.height then
					local dist = the.player:distanceTo(spr)
					local xdist = 200
					local ydist = 200
					if spr.x <= the.player.x then
						xdist = the.player.x - spr.x
					else
						xdist = spr.x - the.player.x - the.player.width
					end
					if spr.y <= the.player.y then
						ydist = the.player.y - spr.y
					else
						ydist = spr.y - the.player.y - the.player.height
					end
					--the.view.GUI.MiscText.text = "Grab from " .. xdist .. "," .. ydist
					if xdist <= 35 and ydist <= 35 then
						spr.solid = false
						self.SelectedBox = spr
						self.SelectedBox.inHands = true
						self.SelectedBox.alpha = 0.5
						TEsound.play('wav/box_grab.wav', '', 0.6)
					end
				end
			end
		end
	end,
	
	onLeftMouseUp = function (self, xpos, ypos)
	end,
	
	onUpdate = function (self, xpos, ypos)
		if self.SelectedBox then
				self.SelectedBox.velocity.y = 0
			if the.player.facing == 0 then
				self.SelectedBox.x = the.player.x
				self.SelectedBox.y = the.player.y + the.player.width / 2 
			elseif the.player.facing == 1 then
				self.SelectedBox.x = the.player.x + the.player.width + 4
				self.SelectedBox.y = the.player.y 
			elseif the.player.facing == -1 then
				self.SelectedBox.x = the.player.x - self.SelectedBox.width - 4
				self.SelectedBox.y = the.player.y 
			end
		end
	end
}

GreenPortal = Animation:extend
{
	width = 32,
	height = 64,
	image = 'img/teleporter.png',
	sequences = 
	{
		active = { frames = {1, 2, 3, 4}, fps = 25 },
	},
	visible = false,
	
	onUpdate = function (self, elapsed)
		self:play('active')
	end	
}

RedPortal = Animation:extend
{
	width = 32,
	height = 64,
	image = 'img/teleporter.png',
	sequences = 
	{
		active = { frames = {5, 6, 7, 8}, fps = 25 },
	},
	visible = false,
	
	targetPortal = nil,
	
	onUpdate = function (self, elapsed)
		self:play('active')
	end,
	
	onCollide = function (self, other, h, v)
		if self.targetPortal and self.targetPortal.visible and h > 30 then
			other.x = self.targetPortal.x
			other.y = self.targetPortal.y
			if self:distanceTo(the.player) <= 550 then
				TEsound.play('wav/portal_enter.wav')
			end
		end
	end
}

TeleDropper = DocGadget:extend
{
	Icon = 'img/tele-dropper.png',
	GadgetName = 'Tele-Dropper',
	
	onNew = function(self)
		self.ActivePortalExit = GreenPortal:new {}
		the.view:add(self.ActivePortalExit)
		self.ActivePortalEntrance = RedPortal:new {targetPortal = self.ActivePortalExit}
		the.view:add(self.ActivePortalEntrance)
	end,
	
	onLeftMouseClick = function(self, xpos, ypos)
		self.ActivePortalExit.x = the.player.x
		self.ActivePortalExit.y = the.player.y
		self.ActivePortalExit.visible = true
		TEsound.play('wav/portal_drop.wav')
	end,
	
	onRightMouseClick = function(self, xpos, ypos)
		self.ActivePortalEntrance.x = the.player.x
		self.ActivePortalEntrance.y = the.player.y
		self.ActivePortalEntrance.visible = true
		TEsound.play('wav/portal_drop.wav')
	end,
	
	onLeftMouseUp = function (self, xpos, ypos)
	end,
	
	onUpdate = function (self, xpos, ypos, elapsed)
		self.ActivePortalEntrance:collide(the.player)
		self.ActivePortalEntrance:collide(the.view.people)
	end
}

FreezeRay = DocGadget:extend
{
	Icon = 'img/ice-gun.png',
	GadgetName = 'Freeze Ray',
	firing = false,
	lastFire = 1.0,
	fireSpeed = 0.1,
	lastPlay = 1.0,
	playEvery = 0.3,
	
	onLeftMouseClick = function(self, xpos, ypos)
		self.firing = true
	end,
	
	onLeftMouseUp = function (self, xpos, ypos)
		self.firing = false
	end,
	
	onUpdate = function (self, xpos, ypos, elapsed)
		self.lastPlay = self.lastPlay + elapsed
		self.lastFire = self.lastFire + elapsed
		if self.firing then
			if (self.lastFire >= self.fireSpeed) then
				local xdir = xpos - the.player.x
				local ydir = ypos - the.player.y
				local xorigin = the.player.x
				local yorigin = the.player.y + the.player.height / 3
				if (the.player.facing == 1) then
					xorigin = xorigin + the.player.width
				elseif(the.player.facing == -1) then
				
				else
					xorigin = xorigin + the.player.width / 4
				end
				
				local blast = FreezeRayBlast:new{x = xorigin, y = yorigin}
				
				local length = math.sqrt(xdir * xdir + ydir * ydir)
				blast.velocity.x = (150 + math.abs(the.player.velocity.x)) * xdir / length
				blast.velocity.y = 150 * ydir / length
				
				
				TEsound.play('wav/ice-gun_shoot.wav')
				the.view.projectiles:add(blast)
				self.lastFire = 0
			end
		end
	end
}

FreezeRayBlast = Tile:extend
{
	image = 'img/tileset.png',
	imageOffset = { x = 0, y = 5*32},
	height = 32,
	width = 32,
	lifeTime = 1.75,
	aliveTime = 0,
	
	onUpdate = function(self, elapsed)
		self.aliveTime = self.aliveTime + elapsed
		if self.aliveTime >= self.lifeTime then
			self:die()
		end
	end,
	
	onCollide = function (self, other, h, v)
		self:die()
	end
}

GUI = Group:extend
{
	Tools = {},
	SelectedTool = 1,
	
	onNew = function (self)
		self.Tools[1] = RobotArms:new{}
		self.Tools[2] = FreezeRay:new{}
		self.Tools[3] = TeleDropper:new{}
	
		self.FPSText = Text:new{x = 0, y = 0, height = 12, width = 800}
		self.MiscText = Text:new{x = 0, y = 12, height = 12, width = 800}
		self.MouseText = Text:new{x = 0, y = 24, height = 12, width = 800}
		self.DebugText = Text:new{x = 0, y = 36, height = 12, width = 800}
		
		self.EnergyBackground = Fill:new { width = 24, height = 450, fill = {175, 175, 0, 100}, border = {255, 255, 0, 255}, x = 724, y = 32 }
		
		self.CurrentTool = Tile:new{image = 'img/robot-arms.png', width = 64, height = 64, x = 704, y = 504}
		
		self:add(self.CurrentTool)
		self:add(self.DebugText)
		self:add(self.FPSText)
		self:add(self.MiscText)
		self:add(self.MouseText)
	end,
	
	Update = function (self, elapsed)
		--self.MouseText.text = "Mouse @ " .. the.mouse.x .. "," ..the.mouse.y
		if the.view.FPS < 58 then
			self.FPSText.text = the.view.FPS .. " FPS"
		end
		
		self.CurrentTool.image = self.Tools[self.SelectedTool].Icon
	end
}

AddGroup = function(g1, g2)
	for _, spr in pairs(g2.sprites) do
		if spr then
			g1:add(spr)
		end
	end
end

PauseScreen = Subview:extend
{
	onNew = function(self)
		drawParent = true
		self.background = Fill:new
		{
			width = 800,
			height = 600,
			x = 0,
			y = 0,
			fill = {0, 0, 0, 100}
		}
		self.MenuButton = Fill:new
		{
			width = 256,
			height = 64,
			x = 272,
			y = 392,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.MenuButtonText = Text:new
		{
			text = 'Main Menu',
			align = 'center',
			font = 32,
			x = 272,
			y = 408,
			width = 256,
			height = 32,
			tint = {0, 0, 0, 255}
		}
		self.ResumeButton = Fill:new
		{
			width = 256,
			height = 64,
			x = 272,
			y = 200,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.ResumeButtonText = Text:new
		{
			text = 'Resume',
			align = 'center',
			font = 32,
			x = 272,
			y = 216,
			width = 256,
			height = 32,
			tint = {0, 0, 0, 255}
		}
		self.RestartButton = Fill:new
		{
			width = 256,
			height = 64,
			x = 272,
			y = 296,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.RestartButtonText = Text:new
		{
			text = 'Restart',
			align = 'center',
			font = 32,
			x = 272,
			y = 312,
			width = 256,
			height = 32,
			tint = {0, 0, 0, 255}
		}
		
		self:add(self.background)
		self:add(self.MenuButton)
		self:add(self.ResumeButton)
		self:add(self.RestartButton)
		self:add(self.MenuButtonText)
		self:add(self.ResumeButtonText)
		self:add(self.RestartButtonText)
	end,
	
	onUpdate = function (self, elapsed)
		if the.mouse.x >= self.MenuButton.x and the.mouse.x <= self.MenuButton.x + self.MenuButton.width and the.mouse.y >= self.MenuButton.y and the.mouse.y <= self.MenuButton.y + self.MenuButton.height then
			self.MenuButton.fill = {170, 170, 170, 255}
			if the.mouse:justPressed('l') then
				TEsound.play('wav/gui_select.wav')
				local oldView = the.app.view
				the.app.view = the.app.MMView
				oldView:die()
			end
		else
			self.MenuButton.fill = {100, 100, 100, 255}
		end
		
		if the.mouse.x >= self.ResumeButton.x and the.mouse.x <= self.ResumeButton.x + self.ResumeButton.width and the.mouse.y >= self.ResumeButton.y and the.mouse.y <= self.ResumeButton.y + self.ResumeButton.height then
			self.ResumeButton.fill = {170, 170, 170, 255}
			if the.mouse:justPressed('l') then
				TEsound.play('wav/gui_select.wav')
				self:deactivate()
			end
		else
			self.ResumeButton.fill = {100, 100, 100, 255}
		end
		
		if the.mouse.x >= self.RestartButton.x and the.mouse.x <= self.RestartButton.x + self.RestartButton.width and the.mouse.y >= self.RestartButton.y and the.mouse.y <= self.RestartButton.y + self.RestartButton.height then
			self.RestartButton.fill = {170, 170, 170, 255}
			if the.mouse:justPressed('l') then
				TEsound.play('wav/gui_select.wav')
				local newView = GameMap:new{mapName = the.currentMap}
				local oldView = the.app.view
				the.app.view = newView
				oldView:die()
			end
		else
			self.RestartButton.fill = {100, 100, 100, 255}
		end
		
		if(the.keys:justPressed('escape')) then
			self:deactivate()
		end
	end
}

NextMapMarker = Tile:extend
{
	mapname = nil
}

GameMap = View:extend
{
	mapName = 'titlemap.lua',
	
	onNew = function (self)
		self.pauseMenu = PauseScreen:new{}
	
		self.people = Group:new {}
		self.mapView = View:new {}
		self.mapView:loadLayers(self.mapName)
		
		self.doors = Group:new {}
		AddGroup(self.doors, self.mapView.doors)
		self.buttons = Group:new {}
		AddGroup(self.buttons, self.mapView.buttons)
		self.switches = Group:new {}
		AddGroup(self.switches, self.mapView.switches)
		self.boxes = Group:new {}
		AddGroup(self.boxes, self.mapView.boxes)
		self.vaults = Group:new {}
		AddGroup(self.vaults, self.mapView.vaults)
		self.signs = Group:new {}
		AddGroup(self.signs, self.mapView.signs)
		
		the.player = Player:new{x = the.docSpawn.x, y = the.docSpawn.y}
		the.player.hasFocus = true
		the.cat = Cat:new{x = the.catSpawn.x, y = the.catSpawn.y}
		
		self.projectiles = Group:new{}
		
		self.focus = the.player
		
		if the.nextmap and self.mapView.nextmap.count == 0 then
			the.nextmap = nil
		end
		
		self:add(self.mapView.background)
		if self.mapView.middleground then
			self:add(self.mapView.middleground)
		end
		self:add(self.vaults)
		self:add(self.signs)
		self:add(self.people)
		self:add(the.player)
		self:add(the.cat)
		self:add(self.boxes)
		self:add(self.switches)
		self:add(self.buttons)
		self:add(self.doors)
		self:add(self.projectiles)
		if self.mapView.foreground then
			self:add(self.mapView.foreground)
		end
		self.GUI = GUI:new{}
		self:add(self.GUI)
		
		self:clampTo(self.mapView.background)
	end,
	
	onStartFrame = function (self)
		
	end,
	
	onUpdate = function (self, elapsed)
		
		self.FPS = math.floor(1 / elapsed)
		
		self.mapView.background:subcollide(self.projectiles)
		self.people:collide(self.projectiles)
		
		self.boxes:collide(self.buttons)
		self.GUI.translate.x = self.translate.x * -1
		self.GUI.translate.y = self.translate.y * -1
		
		
		if the.mouse:justPressed('l') then
			self.GUI.Tools[self.GUI.SelectedTool]:LeftMouseClick()
		end
		if the.mouse:justReleased('l') then
			self.GUI.Tools[self.GUI.SelectedTool]:LeftMouseUp()
		end
		if the.mouse:justPressed('r') then
			self.GUI.Tools[self.GUI.SelectedTool]:RightMouseClick()
		end
		if the.mouse:justReleased('r') then
			self.GUI.Tools[self.GUI.SelectedTool]:RightMouseUp()
		end
		if the.keys:pressed('1') then
			--the.view.GUI.DebugText.text = "Slot 1"
			self.GUI.SelectedTool = 1
		end
		if the.keys:pressed('2') then
			--the.view.GUI.DebugText.text = "Slot 2"
			self.GUI.SelectedTool = 2
		end
		if the.keys:pressed('3') then
			--the.view.GUI.DebugText.text = "Slot 3"
			self.GUI.SelectedTool = 3
		end
		if the.keys:pressed('4') then
			--the.view.GUI.DebugText.text = "Slot 4"
			--self.GUI.CurrentTool = 4
		end
		if the.keys:pressed('5') then
			--the.view.GUI.DebugText.text = "Slot 5"
			--self.GUI.CurrentTool = 5
		end
		if the.keys:justPressed('z') then
			-- switch focus between cat/doc
			if the.player.hasFocus then
				self.focus = the.cat
				the.player.hasFocus = false
				the.cat.hasFocus = true
			elseif the.cat.hasFocus then
				self.focus = the.player
				the.player.hasFocus = true
				the.cat.hasFocus = false
			else
				self.focus = the.player
				the.player.hasFocus = true
				the.cat.hasFocus = false
			end
		end
		if the.keys:justPressed('escape') then
			self.pauseMenu:activate()
		end
		
		self.GUI:Update(elapsed)
		self.GUI.Tools[self.GUI.SelectedTool]:FrameListener(elapsed)
		
		-- check for victory
		self.victory = true
		for _, spr in pairs(self.vaults.sprites) do
			if spr:instanceOf(Vault) and not spr.Opened then
				self.victory = false
			end
		end
		
		if self.victory then
			if the.nextMap and the.nextMap.mapname and the.nextMap.mapname ~= 'none' then
				the.currentMap = the.nextMap.mapname
				the.app.view = GameMap:new{mapName = the.currentMap}
			else
				the.app.view = the.app.MMView
			end
			self:die()
		end
		
	end
}

SelectButton = Group:extend
{
	xo = 0,
	yo = 0,
	text = '',
	
	onNew = function(self)
		self.Background = Fill:new
		{
			x = self.xo,
			y = self.yo,
			height = 20,
			width = 204,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.Text = Text:new
		{
			x = self.xo + 2,
			y = self.yo + 2,
			width = 200,
			height = 16,
			font = 16,
			text = self.text
		}
		self:add(self.Background)
		self:add(self.Text)
	end,
	
	Pushed = function(self)the.currentMap = self.text
		the.app.view = GameMap:new{mapName = the.currentMap}
	end
}

MapSelect = Subview:extend
{
	dir = "",
	mapList = Text:new
	{
		x = 0,
		y = 0,
		width = 800,
		height = 600,
		wordWrap = true
	},
	
	onNew = function (self)
		local files = love.filesystem.enumerate(self.dir)
		
		self.Buttons = {}
		
		self.Background = Fill:new
		{
			x = 0,
			y = 0,
			width = 800,
			height = 600,
			fill = {0, 0, 0, 150 }
		}
		
		self:add(self.Background)
		
		self.ExitText = Text:new
		{
			x = 0,
			y = 600-16,
			font = 12,
			height = 12,
			width = 200,
			text = "Esc to Close"
		}
		
		self:add(self.ExitText)
		local x = 1
		for k, file in pairs(files) do
			if string.sub(file, -3) == 'lua' and file ~= 'main.lua' and file ~= 'TEsound.lua' then
				self.Buttons[x] = SelectButton:new
								  {
									xo = 16 + math.floor(x / 25) * 230,
									yo = 16 + ((x - 1) % 25) * 22,
									text = file
								  }
				self:add(self.Buttons[x])
				x = x + 1
			end
		end
		self.NumButtons = x - 1
	end,
	
	onUpdate = function (self)
		if the.keys:justPressed('escape') then
			self:deactivate()
		else
			for x=1, self.NumButtons do
				if the.mouse.x >= self.Buttons[x].Background.x and the.mouse.x <= self.Buttons[x].Background.x + self.Buttons[x].Background.width and the.mouse.y >= self.Buttons[x].Background.y and the.mouse.y <= self.Buttons[x].Background.y + self.Buttons[x].Background.height then
					self.Buttons[x].Background.fill = {170, 170, 170, 255}
					if the.mouse:justPressed('l') then
						TEsound.play('wav/gui_select.wav')
						self.Buttons[x]:Pushed()
					end
				else
					self.Buttons[x].Background.fill = {100, 100, 100, 255}
				end
			end
		end
	end
}

MainMenu = View:extend
{
	MapList = MapSelect:new{},

	onNew = function (self)
		self.pauseMenu = PauseScreen:new{}
	
		self.people = Group:new {}
		self.mapView = View:new {}
		self.mapView:loadLayers('titlemap.lua')
		
		self.doors = Group:new {}
		AddGroup(self.doors, self.mapView.doors)
		self.buttons = Group:new {}
		AddGroup(self.buttons, self.mapView.buttons)
		self.switches = Group:new {}
		AddGroup(self.switches, self.mapView.switches)
		self.boxes = Group:new {}
		AddGroup(self.boxes, self.mapView.boxes)
		self.vaults = Group:new {}
		AddGroup(self.vaults, self.mapView.vaults)
		self.signs = Group:new {}
		AddGroup(self.signs, self.mapView.signs)
		
		the.player = Player:new{x = the.docSpawn.x, y = the.docSpawn.y}
		--the.player.hasFocus = true
		the.cat = Cat:new{x = the.catSpawn.x, y = the.catSpawn.y}
		
		self.projectiles = Group:new{}
		
		--self.focus = the.player
		
		self:add(self.mapView.background)
		if self.mapView.middleground then
			self:add(self.mapView.middleground)
		end
		self:add(self.vaults)
		self:add(self.signs)
		self:add(self.people)
		self:add(the.player)
		self:add(the.cat)
		self:add(self.boxes)
		self:add(self.switches)
		self:add(self.buttons)
		self:add(self.doors)
		self:add(self.projectiles)
		if self.mapView.foreground then
			self:add(self.mapView.foreground)
		end		
		self:clampTo(self.mapView.background)
		
		self.TitleImage = Tile:new
		{
			image = 'img/Title.png',
			y = 32,
			x = 160
		}
		self.StartButton = Fill:new
		{
			width = 256,
			height = 64,
			x = 272,
			y = 200,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.StartButtonText = Text:new
		{
			text = 'New Game!',
			align = 'center',
			font = 32,
			x = 272,
			y = 216,
			width = 256,
			height = 32,
			tint = {0, 0, 0, 255}
		}
		self.MapButton = Fill:new
		{
			width = 256,
			height = 64,
			x = 272,
			y = 296,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.MapButtonText = Text:new
		{
			text = 'Pick Map',
			align = 'center',
			font = 32,
			x = 272,
			y = 296+16,
			width = 256,
			height = 32,
			tint = {0, 0, 0, 255}
		}
		self.QuitButton = Fill:new
		{
			width = 256,
			height = 64,
			x = 272,
			y = 392,
			fill = {100, 100, 100, 255},
			border = {50, 50, 50, 255}
		}
		self.QuitButtonText = Text:new
		{
			text = 'Quit',
			align = 'center',
			font = 32,
			x = 272,
			y = 392+16,
			width = 256,
			height = 32,
			tint = {0, 0, 0, 255}
		}
		
		--self:add(self.Background)
		self:add(self.TitleImage)
		self:add(self.StartButton)
		self:add(self.MapButton)
		self:add(self.QuitButton)
		self:add(self.StartButtonText)
		self:add(self.MapButtonText)
		self:add(self.QuitButtonText)
	end,
	
	onUpdate = function (self, elapsed)
		if the.mouse.x >= self.StartButton.x and the.mouse.x <= self.StartButton.x + self.StartButton.width and the.mouse.y >= self.StartButton.y and the.mouse.y <= self.StartButton.y + self.StartButton.height then
			self.StartButton.fill = {170, 170, 170, 255}
			if the.mouse:justPressed('l') then
				TEsound.play('wav/gui_select.wav')
				the.currentMap = 'level1.lua'
				the.app.view = GameMap:new{mapName = the.currentMap}
			end
		else
			self.StartButton.fill = {100, 100, 100, 255}
		end
		if the.mouse.x >= self.MapButton.x and the.mouse.x <= self.MapButton.x + self.MapButton.width and the.mouse.y >= self.MapButton.y and the.mouse.y <= self.MapButton.y + self.MapButton.height then
			self.MapButton.fill = {170, 170, 170, 255}
			if the.mouse:justPressed('l') then
				TEsound.play('wav/gui_select.wav')
				self.MapList:activate()
			end
		else
			self.MapButton.fill = {100, 100, 100, 255}
		end
		if the.mouse.x >= self.QuitButton.x and the.mouse.x <= self.QuitButton.x + self.QuitButton.width and the.mouse.y >= self.QuitButton.y and the.mouse.y <= self.QuitButton.y + self.QuitButton.height then
			self.QuitButton.fill = {170, 170, 170, 255}
			if the.mouse:justPressed('l') then
				TEsound.play('wav/gui_select.wav')
				the.app:quit()
			end
		else
			self.QuitButton.fill = {100, 100, 100, 255}
		end
	end
}

the.app = App:new
{
	onNew = function (self)
		TEsound.tagVolume('doorsw', .4)
		TEsound.tagVolume('all', .6)
		self:add(Text:new
		{
			y = 300, width = the.app.width, align = 'center',
		})
			
			
		TEsound.playLooping('wav/bgm.ogg', 'bgm', nil, 0.3)
		the.currentMap = 'titlemap.lua'
		self.MMView= MainMenu:new{}
		self.view = self.MMView
	end,
	
	onUpdate = function (self, elapsed)
		TEsound.cleanup()
	end
}
